
#### 
#
#sampParams = setSamplingParameters(na=19, nb = 17, sig = diag(3) *0.04, mudiff=0.3)
#
####

setSamplingParameters = function(na,sig, nb = NULL, sigb = NULL, mudiff, conditioning = 1) {
  # mudiff is mean difference between group a and group b
  # we set the mean of group b to 0, and group a is mudiff

  # Fill in group b with defaults 
  if (is.null(nb)) { nb = na}
  if (is.null(sigb)) { sigb = sig }
  stopifnot(ncol(sig)==ncol(sigb))

  # expand mudiff
  d = ncol(sig)
  if (length(mudiff) == 1){ mudiff = rep(mudiff,d) }
  stopifnot(length(mudiff) == d )
  
  # set param structure
  sampParam = list(na = na, nb = nb, sig = sig, sigb = sigb, mudiff = mudiff,d = ncol(sig), 
                   conditioning = conditioning)
  return(sampParam)
}

setBlockSamplingParameters = function(na,sigs, nb = NULL, sigbs = NULL, mudiff, conditioning = 1) {
    # similar to "setSamplingParameters" but simulated in independent blocks
    # sigs, sigbs are lists of covariance matrices

    # mudiff is mean difference between group a and group b
    # we set the mean of group b to 0, and group a is mudiff
    
    # Fill in group b with defaults 
    if (is.null(nb)) { nb = na}
    if (is.null(sigbs)) { sigbs = sigs }
    d = sum(sapply(sigs, ncol))
    stopifnot(d == sum(sapply(sigbs, ncol)))
    
    # expand mudiff
    if (length(mudiff) == 1){ mudiff = rep(mudiff,d) }
    stopifnot(length(mudiff) == d )
    
    # set param structure
    sampParam = list(na = na, nb = nb, sigs = sigs, sigbs = sigbs, mudiff = mudiff,d = d, 
                     conditioning = conditioning)
    return(sampParam)
}

#B = 500   # Parametric bootstrap samples

# condSample(B = 500, na = 18, nb = 15, sampleSig = diag(3)*0.04, m=0)
condSample = function(B,sP, thresh, returndat = FALSE){
  require("MASS")
  # The Y matrix is the full data matrix
  # Y_a are samples for group a, Y_b for group b
  # Z is the mean-difference vector
  # cond means conditional on thresholding. 
  # sP$conditioning is a vector: conditioning[i] = 1 means Z[i]>thresh
  #                           conditioning[i] = -1 means Z[i] < thresh
  #                           conditioning[i] = 0 means Z[i] no condition
  #              it is checked using all((conditioning * Z) >= (conditioning * thresh)
  
  # Simulate the conditional distribution 

  groups = rep(c(1,0),times = c(sP$na,sP$nb))

  condZ = matrix(nr = B, nc = sP$d)
  if (returndat){
    condY = array(0, c(sP$na+sP$nb,sP$d,B))
  } else{
    condY = NULL
  }
  groups = c(rep(1,sP$na),rep(0,sP$nb))

  if (is.null( sP$conditioning)){
    conditioning = 1
  } else {
    conditioning = sP$conditioning
  }
  stopifnot(is.element(length(conditioning), c(1, sP$d)))
  
  B_i = 0   # counter passed conditioning
  i = 0     # global counter
  while(B_i<B){
    i = i + 1
    Y_a = mvrnorm(n = sP$na,mu = sP$mudiff,Sigma = sP$sig)
    Y_b = mvrnorm(n = sP$nb,mu = rep(0, sP$d),Sigma = sP$sigb)
    Z = colMeans(Y_a) - colMeans(Y_b)

    # check condition:
    check_cond = all(Z*conditioning >= thresh * conditioning)
    if (check_cond) {
      B_i = B_i + 1
      condZ[B_i,] = Z 
      if (returndat){
        condY[,,B_i] = rbind(Y_a, Y_b)
      }
    }
  }
  return(list(condZ = condZ, samples =  i, condY= condY, groups = groups))
}



blockCondSample = function(B,sP_blocks, thresh, returndat = FALSE){
    require("MASS")
    # The Y matrix is the full data matrix
    # Y_a are samples for group a, Y_b for group b
    # Z is the mean-difference vector
    # cond means conditional on thresholding. 
    # sP$conditioning is a vector: conditioning[i] = 1 means Z[i]>thresh
    #                           conditioning[i] = -1 means Z[i] < thresh
    #                           conditioning[i] = 0 means Z[i] no condition
    #              it is checked using all((conditioning * Z) >= (conditioning * thresh)
    
    # Simulate the conditional distribution 
    
    groups = rep(c(1,0),times = c(sP_blocks$na,sP_blocks$nb))
    
    condZ = matrix(nr = B, nc = sP_blocks$d)
    if (returndat){
        condY = array(0, c(sP_blocks$na+sP_blocks$nb,sP_blocks$d,B))
    } else{
        condY = NULL
    }
    groups = c(rep(1,sP_blocks$na),rep(0,sP_blocks$nb))
    
    if (is.null( sP_blocks$conditioning)){
        conditioning = 1
    } else {
        conditioning = sP_blocks$conditioning
    }
    stopifnot(is.element(length(conditioning), c(1, sP_blocks$d)))
    
    # We simulate block by block...
    blocks = length(sP_blocks$sigs)
    block_sizes = sapply(sP_blocks$sigs,ncol)

    block_ends = cumsum(block_sizes)
    block_starts = c(1,block_ends+1)[1:blocks] 
        
    for (bl in 1:blocks){
        B_i = 0   # counter passed conditioning
        i = 0     # global counter
        bl_s = block_starts[bl]
        bl_e = block_ends[bl]
        bl_d = block_sizes[bl]
        while(B_i<B){
            i = i + 1
            Y_a = mvrnorm(n = sP_blocks$na,mu = sP_blocks$mudiff[bl_s:bl_e],
                          Sigma = sP_blocks$sigs[[bl]])
            Y_b = mvrnorm(n = sP_blocks$nb,mu = rep(0, bl_d),
                          Sigma = sP_blocks$sigbs[[bl]])
            Z = colMeans(Y_a) - colMeans(Y_b)
        
            #check condition:
            check_cond = all(Z*conditioning[bl_s:bl_e] >= 
                                                thresh * conditioning[bl_s:bl_e])
            if (check_cond) {
                B_i = B_i + 1
                condZ[B_i,bl_s:bl_e] = Z 
                if (returndat){
                    condY[,bl_s:bl_e,B_i] = rbind(Y_a, Y_b)
                }
            }
        }
    }
    return(list(condZ = condZ, samples =  i, condY= condY, groups = groups))
}

diffCov = function(sampleCov, groups) {
    n_a = sum(groups==0)
    n_b = sum(groups==1)
    return(sampleCov * (1/n_a + 1/n_b) )
}

covFromBlocks = function(sigs){
    blocks = length(sigs)
    block_sizes = sapply(sigs,ncol)
    
    block_ends = cumsum(block_sizes)
    block_starts = c(1,block_ends+1)[1:blocks] 
    
    d = sum(block_sizes)
    
    cov_mat = matrix(0,nc = d,nr = d)
    for (i in 1:blocks){
        rg = block_starts[i]:block_ends[i]
        cov_mat[rg,rg] = sigs[[i]]
    }
    return(cov_mat)
}